function CPlayer:GetDownwardCollisions( map, y1, y2 )
--
	-- now we need the tile positions on the map
	local cy1 = floor( y1 / map.tileHeight );
	local cy2 = floor( y2 / map.tileHeight );
	
	-- even though we're considering downward movement
	-- we have to cover all the tiles the sprite takes up
	-- plus the tiles the movement takes up.  So we also
	-- need the horizontal tiles.
	local x1 = self:GetXPos();
	local x2 = ( self:GetXPos() + self:GetWidth() ) - 1;
	
	local cx1 = floor( x1 / map.tileWidth );
	local cx2 = floor( x2 / map.tileWidth );
	
	local floorPos = nil;
	
	for loopCy = cy1, cy2 do
	--
		floorPos = nil;
		for loopCx = cx1, cx2 do
		--
			tile = map( "Tile Layer 1" )( loopCx, loopCy );
			if( tile )then
			--
				local isBlocked = tile.properties.block_n;
				local tileType = tile.properties.type;
				if( isBlocked or tileType == "border" )then
				--
					-- this bit just makes sure we get the highest tile position
					local tileTop = loopCy * map.tileHeight;
					if( not( floorPos ) )then
						floorPos = tileTop;
					elseif( floorPos > tileTop )then
						floorPos = tileTop;
					end;
				--
				elseif( tileType == "slope" )then

					local xOverlap = nil;
					
					-- checking for flipped tiles tells which side we should bias the x overlap on
					local isFlipped = IsMapLayerTileFlippedHorz( map( "Tile Layer 1" ), loopCx, loopCy );
					if( not( isFlipped ) )then
						-- how far into the current tile we are.  Since we subtracted 1 from 
						-- the width, we need to add it back in for the proper overlap
						xOverlap = floor( x2 - ( loopCx * map.tileWidth ) ) + 1;
						if( xOverlap > 16 )then	xOverlap = 16; end;						
					else
						xOverlap = floor( ( ( loopCx * map.tileWidth ) + map.tileWidth ) - x1 ) + 1;
						if( xOverlap > 16 )then	xOverlap = 16; end;						
					end;
						
					local floorHeight = nil;

					-- next up we want the heightmap of the tile
					local heights = fromCSV( tile.properties.heightmap );
					if( heights )then
						floorHeight = heights[ xOverlap ];
					end;

					if( floorHeight )then
						local tileBase = ( ( loopCy * map.tileHeight ) + map.tileHeight );
						local floorTop = tileBase - floorHeight;

                                                if( y2 > floorTop )then
                                                        if( not( floorPos ) )then
                                                                floorPos = floorTop;
                                                        elseif( floorPos > floorTop )then
                                                                floorPos = floorTop;
                                                        end;
                                                end;						
					end;

				end; -- if tileType
			--
			end; -- if tile
		--
		end; -- for loopCx
		
		if( floorPos )then
			break;
		end;
	--
	end; -- for loopCy
	
	return( floorPos );
--
end;

