CPlayer = {};

function CPlayer:new()
--
	local object = 
	{
		rect = CRect:new(),
		bounds = CRect:new(),
		speed = CPoint:new(),
		velocity = CPoint:new(),
		minVelocity = CPoint:new(),
		maxVelocity = CPoint:new(),
		acc = 0,
		dec = 0,
		jumpForce = 0,
		angle = 0,
		groundSpeed = 0,
		maxGroundSpeed = 0,
		collisionPoints = { CPoint:new(), CPoint:new(), 
			CPoint:new(), CPoint:new(), CPoint:new(), CPoint:new() },
		facing = "right",
		groundType = "none",
		state = "falling"
	};
	
	setmetatable( object, { __index = CPlayer } );
	
	return object;
--
end;

function CPlayer:SetPos(x, y)
--
	self.rect:SetPos(x, y);
	
	local w = self:GetWidth();
	local h = self:GetHeight();
	self.collisionPoints[ CP_RIGHT ]:Set( ( x + w ) - 1, ( y + h ) - 1 );
	self.collisionPoints[ CP_LEFT ]:Set( x, ( y + h ) - 1 );
	self.collisionPoints[ CP_BOTTOM_RIGHT ]:Set( ( x + w ) - 1, ( y + h ) - 1 );
	self.collisionPoints[ CP_BOTTOM_LEFT ]:Set( x, ( y + h ) - 1 );
	self.collisionPoints[ CP_TOP_LEFT ]:Set( x, y );
	self.collisionPoints[ CP_TOP_RIGHT ]:Set( ( x + w ) - 1, y );
	
	self:KeepInBounds();
--
end;

function CPlayer:SetXPos( x )
--
	self.rect:SetXPos( x );
	-- I want the point to be on the right side
	local w = self:GetWidth();
	self.collisionPoints[ CP_RIGHT ]:SetX( ( x + w ) - 1 );
	self.collisionPoints[ CP_LEFT ]:SetX( x );
	self.collisionPoints[ CP_BOTTOM_RIGHT ]:SetX( ( x + w ) - 1 );
	self.collisionPoints[ CP_BOTTOM_LEFT ]:SetX( x );
	self.collisionPoints[ CP_TOP_LEFT ]:SetX( x );
	self.collisionPoints[ CP_TOP_RIGHT ]:SetX( ( x + w ) - 1 );
	
	self:KeepInHorzBounds();
--
end;

function CPlayer:SetYPos( y )
--
	self.rect:SetYPos( y );
	
	local h = self:GetHeight();
	self.collisionPoints[ CP_RIGHT ]:SetY( ( y + h ) - 1 );
	self.collisionPoints[ CP_LEFT ]:SetY( ( y + h ) - 1 );
	self.collisionPoints[ CP_BOTTOM_RIGHT ]:SetY( ( y + h ) - 1 );
	self.collisionPoints[ CP_BOTTOM_LEFT ]:SetY( ( y + h ) - 1 );
	self.collisionPoints[ CP_TOP_LEFT ]:SetY( y );
	self.collisionPoints[ CP_TOP_RIGHT ]:SetY( y );

	self:KeepInVertBounds();
--
end;

function CPlayer:GetPos()
--
	return self.rect:GetPos();		
--
end;

function CPlayer:GetXPos()
--
	return self.rect:GetXPos();
--
end;

function CPlayer:GetYPos()
--
	return self.rect:GetYPos();
--
end;

function CPlayer:SetSize(w, h)
--
	self.rect:SetSize( w, h );
	
	local x = self:GetXPos();
	local y = self:GetYPos();
	local w = self:GetWidth();
	local h = self:GetHeight();
	self.collisionPoints[ CP_RIGHT ]:Set( ( x + w ) - 1, ( y + h ) - 1 );
	self.collisionPoints[ CP_LEFT ]:Set( x, ( y + h ) - 1 );
	self.collisionPoints[ CP_BOTTOM_RIGHT ]:Set( ( x + w ) - 1, ( y + h ) - 1 );
	self.collisionPoints[ CP_BOTTOM_LEFT ]:Set( x, ( y + h ) - 1 );
	self.collisionPoints[ CP_TOP_LEFT ]:Set( x, y );
	self.collisionPoints[ CP_TOP_RIGHT ]:Set( ( x + w ) - 1, y );
--
end;

function CPlayer:SetWidth( w )
--
	self.rect:SetWidth( w );
	
	-- I want the point to be on the right side
	local x = self:GetXPos();
	local w = self:GetWidth();
	self.collisionPoints[ CP_RIGHT ]:SetX( ( x + w ) - 1 );
	self.collisionPoints[ CP_LEFT ]:SetX( x );
	self.collisionPoints[ CP_BOTTOM_RIGHT ]:SetX( ( x + w ) - 1 );
	self.collisionPoints[ CP_BOTTOM_LEFT ]:SetX( x );
	self.collisionPoints[ CP_TOP_LEFT ]:SetX( x );
	self.collisionPoints[ CP_TOP_RIGHT ]:SetX( ( x + w ) - 1 );
--
end;

function CPlayer:SetHeight( h )
--
	self.rect:SetHeight( h );
	-- I want the point to be about 3/4ths of the way down.
	local y = self:GetYPos();
	local h = self:GetHeight();
	self.collisionPoints[ CP_RIGHT ]:SetY( ( y + h ) - 1 );
	self.collisionPoints[ CP_LEFT ]:SetY( ( y + h ) - 1 );
	self.collisionPoints[ CP_BOTTOM_RIGHT ]:SetY( ( y + h ) - 1 );
	self.collisionPoints[ CP_BOTTOM_LEFT ]:SetY( ( y + h ) - 1 );
	self.collisionPoints[ CP_TOP_LEFT ]:SetY( y );
	self.collisionPoints[ CP_TOP_RIGHT ]:SetY( y );
--
end;

function CPlayer:GetSize()
--
	return self.rect:GetSize();		
--
end;

function CPlayer:GetWidth()
--
	return self.rect:GetWidth();
--
end;

function CPlayer:GetHeight()
--
	return self.rect:GetHeight();
--
end;

function CPlayer:GetCollisionPoint( pt )
--
	return self.collisionPoints[ pt ]:GetPt();
--
end;

function CPlayer:GetCollisionPointX( pt )
--
	return self.collisionPoints[ pt ]:GetX();
--
end;

function CPlayer:GetCollisionPointY( pt )
--
	return self.collisionPoints[ pt ]:GetY();
--
end;

function CPlayer:SetBounds( x, y, w, h )
--
	self.bounds:SetPos( x, y );
	self.bounds:SetSize( w, h );
--
end;

function CPlayer:KeepInHorzBounds()
--
	local width = self:GetWidth();
	local left = self:GetXPos();
	local right = left + width;

	local left_bound = self.bounds:GetXPos();
	local right_bound = left_bound + self.bounds:GetWidth();
	
	if( left < left_bound )then
		self.rect:SetXPos( left_bound );
	end;
		
	if( right > right_bound )then
		self.rect:SetXPos( right_bound - width );
	end;
--
end;

function CPlayer:KeepInVertBounds()
--
	local height = self:GetHeight();
	local top = self:GetYPos();
	local bottom = top + height;
	local top_bound = self.bounds:GetYPos();
	local bottom_bound = top_bound + self.bounds:GetHeight();
	
	if( top < top_bound )then
		self.rect:SetYPos( top_bound );
	end;
	
	if( bottom > bottom_bound )then
		self.rect:SetYPos( bottom_bound - height );
	end;
--
end;

function CPlayer:KeepInBounds()
--
	self:KeepInHorzBounds();
	self:KeepInVertBounds();
--
end;

function CPlayer:SetGroundSpeed( speed )
--
	self.groundSpeed = speed;
--
end;

function CPlayer:SetMaxGroundSpeed( maxSpeed )
--
	self.maxGroundSpeed = maxSpeed;
--
end;

function CPlayer:SetAcceleration( acc )
--
	self.acc = acc;
--
end;

function CPlayer:SetDeceleration( dec )
--
	self.dec = dec;
--
end;

function CPlayer:SetJumpForce( jumpForce )
--
	self.jumpForce = jumpForce;
--
end;

function CPlayer:SetMinVelocity( xVel, yVel )
--
	self.minVelocity:Set( xVel, yVel );
--
end;

function CPlayer:SetMaxVelocity( xVel, yVel )
--
	self.maxVelocity:Set( xVel, yVel );
--
end;

function CPlayer:SetVelocity(x, y)
--
	local xVel = x;
	if( xVel < self.minVelocity:GetX() )then
		xVel = self.minVelocity:GetX();
	elseif( xVel > self.maxVelocity:GetX() )then
		xVel = self.maxVelocity:GetX();
	end;

	local yVel = y;
	if( yVel < self.minVelocity:GetY() )then
		yVel = self.minVelocity:GetY();
	elseif( yVel > self.maxVelocity:GetY() )then
		yVel = self.maxVelocity:GetY();
	end;
	
	self.velocity:Set( xVel, yvel );	
--
end;

function CPlayer:SetXVelocity( x )
--
	local xVel = x;
	if( xVel < self.minVelocity:GetX() )then
		xVel = self.minVelocity:GetX();
	elseif( xVel > self.maxVelocity:GetX() )then
		xVel = self.maxVelocity:GetX();
	end;

	self.velocity:SetX( xVel );
--
end;

function CPlayer:SetYVelocity( y )
--
	local yVel = y;
	if( yVel < self.minVelocity:GetY() )then
		yVel = self.minVelocity:GetY();
	elseif( yVel > self.maxVelocity:GetY() )then
		yVel = self.maxVelocity:GetY();
	end;

	self.velocity:SetY( yVel );
--
end;

function CPlayer:GetVelocity()
--
	return self.velocity:Get();		
--
end;

function CPlayer:GetXVelocity()
--
	return self.velocity:GetX();
--
end;

function CPlayer:GetYVelocity()
--
	return self.velocity:GetY();
--
end;

function CPlayer:GenerateMovementCommands( isLeftArrowDown, isRightArrowDown )
--
	
	local angleInRadians = math.rad( self.angle );
	local cosOfAngleInRadians = math.cos( angleInRadians );
	
	local maxGSpd = self.maxGroundSpeed;
	if( self.groundType == "downhill_left" or self.groundType == "downhill_right" )then
		maxGSpd = self.maxGroundSpeed * 2;
	elseif( self.groundType == "uphill_left" or self.groundType == "uphill_right" )then
		maxGSpd = self.maxGroundSpeed / 2;
	end;
	
	if ( isLeftArrowDown ) then
	--
		if( self.groundSpeed > 0 )then
			self.groundSpeed = 0;
		end;
		
		self.groundSpeed = self.groundSpeed - self.acc;
				
		if( self.groundSpeed < -maxGSpd )then
			self.groundSpeed = -maxGSpd;
		end;		
		self.facing = "left";
	--
	elseif ( isRightArrowDown ) then
	--
		if( self.groundSpeed < 0 )then
			self.groundSpeed = 0;
		end;
		self.groundSpeed = self.groundSpeed + self.acc;
		if( self.groundSpeed > maxGSpd )then
			self.groundSpeed = maxGSpd;
		end;		
		self.facing = "right";
	--
	else
	--
		if( self.facing == "right" and self.groundSpeed > 0 )then
			self.groundSpeed = self.groundSpeed - self.dec;
			if( self.groundSpeed < 0 )then
				self.groundSpeed = 0;
			end;		
		elseif( self.facing == "left" and self.groundSpeed < 0 )then
			self.groundSpeed = self.groundSpeed + self.dec;
			if( self.groundSpeed > 0 )then
				self.groundSpeed = 0;
			end;		
		end;
	--
	end;	

	if( self.groundSpeed ~= 0 )then
		self:SetXVelocity( self.groundSpeed * cosOfAngleInRadians );
	else
		self:SetXVelocity( 0 );
	end;	
--
end;

function CPlayer:AdjustY( heights, xPos, cx, cy, tileIsFlipped )
--
	if( heights )then
	--
		local floorHeight = nil;
		
		if( not( tileIsFlipped ) )then
			-- left side of the tile we're working with
			local tileLeft = cx * map.tileWidth;
			-- how much the move covers the tile (loopX)
			local xOverlap = ( floor( xPos ) - tileLeft ) + 1;
			if( xOverlap > 16 )then xOverlap = 16; end;
			floorHeight = heights[ xOverlap ];
		elseif( tileIsFlipped )then
			local tileRight = ( cx * map.tileWidth ) + map.tileWidth;
			local xOverlap = ( tileRight - floor( xPos ) );
			if( xOverlap > 16 )then xOverlap = 16; end;
			floorHeight = heights[ xOverlap ];
		end;
		
		if( floorHeight )then
			local tileBottom = ( cy * map.tileHeight ) + map.tileHeight;
			local floorPos = tileBottom - floorHeight;
			self:SetYPos( floorPos  - ( self:GetHeight() ) );
		end;
	--
	end;
--
end;

function CPlayer:MoveLeft( map, x1, x2 )
--
	local cellY = 0;
	local latestLegalXPos = nil;
	local inc = 1;

	local yp = self:GetCollisionPointY( CP_LEFT );
	cellY = floor( yp / map.tileHeight );
	
	local cx1 = floor( x1 / map.tileWidth );
	local cx2 = floor( x2 / map.tileWidth );

	for loopX = cx1, cx2, -inc do
	--	
		local tile = map("Tile Layer 1")(loopX, cellY);
		if ( tile ) then
			local tileType = tile.properties.type;
			local isBlocked = tile.properties.block_e;
			if( tileType == "border" or isBlocked )then
				latestLegalXPos = ( loopX * map.tileWidth ) + map.tileWidth;
				self:SetXVelocity( 0 );
				self:SetGroundSpeed( 0 );
				break;
			else
				latestLegalXPos = x2;

				if( tileType == "slope" )then
					heights = fromCSV( tile.properties.heightmap );
					if( heights )then
						self.angle = tile.properties.angle;
						local isTileFlipped = IsMapLayerTileFlippedHorz( map("Tile Layer 1"), loopX, cellY );
						if( not( isTileFlipped ) )then
							self.groundType = "downhill_left";
						else
							self:AdjustY( heights, x2, loopX, cellY, isTileFlipped );
							self.groundType = "uphill_left";
						end;
					end;
				else
					latestLegalXPos = x2;
				end;				
			end;
		else
			latestLegalXPos = x2;
		end;
	--	
	end;

	if( latestLegalXPos )then
		self:SetXPos( latestLegalXPos );
	end;
--
end;

function CPlayer:MoveRight( map, x1, x2 )
--
	local latestLegalXPos = nil;

	local yp = self:GetCollisionPointY( CP_RIGHT );
	local cellY = floor( yp / map.tileHeight );
	
	local inc = 1;

	local cx1 = floor( x1 / map.tileWidth );
	local cx2 = floor( x2 / map.tileWidth );

	for loopX = cx1, cx2, inc do
	--			
		local tile = map("Tile Layer 1")(loopX, cellY);
		if ( tile ) then
			local tileType = tile.properties.type;
			local isBlocked = tile.properties.block_w;
			if( tileType == "border" or isBlocked )then
				latestLegalXPos = ( loopX * map.tileWidth ) - 1;
				self:SetXVelocity( 0 );
				self:SetGroundSpeed( 0 );
				break;
			else
				latestLegalXPos = x2;

				if( tileType == "slope" )then
					heights = fromCSV( tile.properties.heightmap );
					if( heights )then
						local isTileFlipped = IsMapLayerTileFlippedHorz( map("Tile Layer 1"), loopX, cellY );
						self.angle = tile.properties.angle;
						if( not( isTileFlipped ) )then
							self:AdjustY( heights, x2, loopX, cellY, isTileFlipped );
							self.groundType = "uphill_right";
						else
							self.groundType = "downhill_right";
						end;
					end;
				else
					latestLegalXPos = x2;
				end;
			end;
		else
			latestLegalXPos = x2;
		end;
	--
	end;
	
	if( latestLegalXPos )then
		self:SetXPos( latestLegalXPos  - ( self:GetWidth() - 1 ) );		
	end;
--
end;

function CPlayer:UpdateHorz( dt, map )
--
	local xVel = self:GetXVelocity();
	
	local dx = self:GetXVelocity() * dt;
	
	if( dx > 0 )then
		local xp = self:GetCollisionPointX( CP_RIGHT );
		self:MoveRight( map, xp, xp + dx );
	elseif( dx < 0 )then
		local xp = self:GetCollisionPointX( CP_LEFT );
		self:MoveLeft( map, xp, xp + dx );
	end;
--
end;

function CPlayer:GetNextValidYPosMovingUp( map, pt, dy )
--
	local cellX = 0;
	local latestLegalYPos = nil;
	local inc = 1;

	local yp = pt:GetY();
	local yt = yp;
	local yb = yp + dy;
	
	local cy1 = floor( yt / map.tileHeight );
	local cy2 = floor( yb / map.tileHeight );
	
	for loopY = cy1, cy2, -inc do
	--	
		local xp = pt:GetX();
		cellX = floor( xp / map.tileWidth );

		local tile = map( "Tile Layer 1" )( cellX, loopY );
		if ( tile ) then
			local tileType = tile.properties.type;
			local isBlocked = tile.properties.block_s;
			if( tileType == "border" or isBlocked )then
				latestLegalYPos = (loopY * map.tileHeight) + map.tileHeight;
				self:SetYVelocity( 0 );
				break;
			else
				latestLegalYPos = yb;
			end;
		else
			latestLegalYPos = yb;
		end;
	--
	end;

	return( latestLegalYPos );
--
end;

function CPlayer:Fly( map, dy )
--
	local lp = self:GetCollisionPoint( CP_TOP_LEFT );
	local rp = self:GetCollisionPoint( CP_TOP_RIGHT );
	
	local ceilLeft = self:GetNextValidYPosMovingUp( map, lp, dy );
	local ceilRight = self:GetNextValidYPosMovingUp( map, rp, dy );
	
	if( ceilLeft and ceilRight )then
		self:SetYPos( math.max( ceilLeft, ceilRight ) );
	elseif( not( ceilRight ) )then
		self:SetYPos( ceilLeft );
	elseif( not( ceilLeft ) )then
		self:SetYPos( ceilRight );
	end;
--
end;

function CPlayer:GetFloorPos( heights, xPos, yPos, cx, cy, tileIsFlipped )
--
	local retValue = nil;

	local floorPos = nil;
	local tileBottom = ( cy * map.tileHeight ) + ( map.tileHeight - 1 );
	
	if( heights )then
	--
		local floorHeight = nil;
		
		if( not( tileIsFlipped ) )then -- slope going up to the right
		--
			local tileLeft = cx * map.tileWidth;
			local xOverlap = ( floor( xPos ) - tileLeft ) + 1;
			if( xOverlap > 16 )then xOverlap = 16; end;
			floorHeight = heights[ xOverlap ];
		--
		elseif( tileIsFlipped )then -- slope going down to the left
		--
			local tileRight = ( cx * map.tileWidth ) + map.tileWidth;
			local xOverlap = ( tileRight - floor( xPos ) );
			if( xOverlap > 16 )then xOverlap = 16; end;
			
			floorHeight = heights[ xOverlap ];
		--	
		end;
		
		if( floorHeight )then
			floorPos = tileBottom - floorHeight;
			if( yPos >= floorPos )then
				retValue = floorPos;
			end;
		end;
	--
	end;

	return( retValue );	
--
end;

function CPlayer:GetNextValidYPosMovingDown( map, pt, dy, dir )
--		
	local latestValidYPos = nil;
	local x1 = pt:GetX();
	local cellX1 = floor( x1 / map.tileWidth );
	local y1 = pt:GetY();
	local y2 = y1 + dy;

	local cy1 = floor(y1 / map.tileHeight);
	local cy2 = floor(y2 / map.tileHeight);		
	local inc = 1;

	local hasCollided = false;
	
	self.angle = 0;

	for loopY = cy1, cy2, inc do
	--
		local tile = map("Tile Layer 1")(cellX1, loopY);
		
		if ( tile ) then
		--
			local tileType = tile.properties.type;
			
			local isBlocked = tile.properties.block_n;
			
			if( tileType == "border" or isBlocked )then
				self.angle = 0;
				latestValidYPos = ( loopY * map.tileHeight ) - 1;
				self:SetYVelocity(0);
				self.groundType = "level";
				hasCollided = true;
				break;
			else

				if( tileType == "slope" )then
				--
					local heights = fromCSV( tile.properties.heightmap );
					local isTileFlipped = IsMapLayerTileFlippedHorz( map("Tile Layer 1"), cellX1, loopY );

					self.angle = tile.properties.angle;
					if( ( dir == "left" and isTileFlipped ) or ( dir == "right" and not( isTileFlipped ) ) )then
						latestValidYPos = self:GetFloorPos( heights, x1, y2, cellX1, loopY, isTileFlipped );
					end;
					
					if( latestValidYPos )then
						hasCollided = true;
						self:SetYVelocity(0);
						
						if( self.facing == "right" and not( isTileFlipped ) )then
							self.groundType = "uphill_right";
						elseif( self.facing == "right" and isTileFlipped )then
							self.groundType = "downhill_right";
						elseif( self.facing == "left" and not( isTileFlipped ) )then
							self.groundType = "downhill_left";
						elseif( self.facing == "left" and isTileFlipped )then
							self.groundType = "uphill_left";
						end;
						break;
					else
						latestValidYPos = y2;
					end;
				--
				else
					latestValidYPos = y2;
				end;
			end;
		--
		else		
			latestValidYPos = y2;
		end;
	--
	end;

	--[[
	if(dir=="right")then
	print(self.angle);
	print(y1);
	print(y2);
	print(cy1);
	print(cy2);
	print(self:GetYVelocity());
	print(hasCollided);
	print(x1);
	print(cellX1);
	print("---");
	end;
	--]]
	return( latestValidYPos );
--
end;

function CPlayer:Fall( map, dy )
--
	local leftPt = self:GetCollisionPoint( CP_BOTTOM_LEFT );		
	local rightPt = self:GetCollisionPoint( CP_BOTTOM_RIGHT );
	local leftPtAngle = self.angle;
	local rightPtAngle = self.angle;

	local latestYPos1 = self:GetNextValidYPosMovingDown( map, leftPt, dy, "left" );
	leftPtAngle = self.angle;

	local latestYPos2 = self:GetNextValidYPosMovingDown( map, rightPt, dy, "right" );
	rightPtAngle = self.angle;

	if( self.facing == "left" )then
		if( leftPtAngle > 0 )then
			self.angle = leftPtAngle;
		else
			self.angle = rightPtAngle;
		end;
	elseif( self.facing == "right" )then
		if( rightPtAngle > 0 )then
			self.angle = rightPtAngle;
		else
			self.angle = leftPtAngle;
		end;
	else
		self.angle = 0;
	end;

	if( latestYPos1 and latestYPos2 )then
		self:SetYPos( math.min( latestYPos1, latestYPos2 )  - ( self:GetHeight() - 1 ) );
	elseif( not( latestYPos1 ) )then
		self:SetYPos( latestYPos2  - ( self:GetHeight() - 1 ) );
	elseif( not( latestYPos2 ) )then
		self:SetYPos( latestYPos1  - ( self:GetHeight() - 1 ) );
	end;

	--print(self.angle);
	--print("---");
--
end;

function CPlayer:ManageGroundState()
--
	local leftPoint = self:GetCollisionPoint( CP_BOTTOM_LEFT );		
	local rightPoint = self:GetCollisionPoint( CP_BOTTOM_RIGHT );

	local lx, ly = leftPoint:Get();
	local rx, ry = rightPoint:Get();
	local cx1 = floor( lx / map.tileWidth );
	local cy1 = floor( ly / map.tileHeight );
	local cx2 = floor( rx / map.tileWidth );
	local cy2 = floor( ry / map.tileHeight );
	
	for loopty = cy1, cy1 + 1 do
		tileL = map( "Tile Layer 1" )( cx1, loopty );
		if( tileL )then
			if( tileL.properties.type == "block" or tileL.properties.type == "border" )then
				self.state = "on_ground";
			elseif( tileL.properties.type == "slope" )then
				local heights = fromCSV( tileL.properties.heightmap );
				local isTileFlipped = IsMapLayerTileFlippedHorz( map("Tile Layer 1"), cx1, loopty );
				local floorPos = nil;			
				for loopy = ly, ly + 8 do
					floorPos = self:GetFloorPos( heights, lx, loopy, cx1, loopty, isTileFlipped );
					if( floorPos )then
						self.state = "on_ground";
						break;
					end;
				end;
				if( floorPos )then break; end;
			end;
		end;
	end;
	
	for loopty = cy2, cy2 + 1 do
		tileR = map( "Tile Layer 1" )( cx2, loopty );
		if( tileR )then
			if( tileR.properties.type == "block" or tileR.properties.type == "border" )then
				self.state = "on_ground";
			elseif( tileR.properties.type == "slope" )then
				local heights = fromCSV( tileR.properties.heightmap );
				local isTileFlipped = IsMapLayerTileFlippedHorz( map("Tile Layer 1"), cx2, loopty );		
				local floorPos = nil;			
				for loopy = ry, ry + 8 do
					floorPos = self:GetFloorPos( heights, rx, loopy, cx2, loopty, isTileFlipped );
					if( floorPos )then
						self.state = "on_ground";
						break;
					end;
				end;
				if( floorPos )then break; end;
			end;
		end;
	end;
	
	if( not( tileL ) and not( tileR ) )then
		self.state = "falling";
	end;
--
end;

function CPlayer:UpdateVert( dt, map, gravity )
--
	local yVel = self:GetYVelocity();
	local angleInRadians = math.rad( self.angle );
	local invSinOfAngleInRadians = -math.sin( angleInRadians );
	local gSpeed = self.groundSpeed;
	if( gSpeed < 0 )then gSpeed = -gSpeed; end;
	local angleVel = -( gSpeed * invSinOfAngleInRadians );
	--if( self.state == "on_ground" )then
		self:SetYVelocity( yVel + angleVel );
	--end;

	local yVel = self:GetYVelocity();
	self:SetYVelocity( yVel + ( gravity * dt) );
	
	-- all this simply means is yVel = yVel + (gravity * dt), then dy = yVel * dt;
	yVel = self:GetYVelocity();
	local dy = yVel * dt;
	
	-- moving down
	if( dy < 0 )then
		self:Fly( map, dy );
		self.groundType = "none";
	elseif( dy > 0 )then
		self:Fall( map, dy );
		self:ManageGroundState();
	end;	
--
end;

function CPlayer:Update( dt, map, gravity )
--
	self:UpdateHorz( dt, map );
	self:UpdateVert( dt, map, gravity );
	self:KeepInBounds();
--
end;

function CPlayer:draw()
--
	local x = floor( self.rect:GetXPos() );
	local y = floor( self.rect:GetYPos() );
	
	love.graphics.rectangle("fill", x, y, self:GetWidth(), self:GetHeight() );
--
end;

--[ [
function CPlayer:Jump(dt, isKeyDown)
--
	-- isKeyDown is null
	if ( isKeyDown ) then
		if ( not ( jump_pressed ) and self.state == "on_ground" ) then
			jump_pressed = true;
			do_jump = true;
		end;
	else
		pressed_time = 0;
		jump_pressed = false;
		do_jump = false;
	end;
	
	if( do_jump ) then
		pressed_time = 
			self:DoJump( dt, pressed_time, MAX_PRESSED_TIME );
	end;
--
end;

function CPlayer:DoJump( dt, air_time, max_air_time )
--
	if ( air_time < max_air_time ) then
		local jump_remaining = ( max_air_time - air_time );

		local jump_time = 0;

		if ( dt > jump_remaining ) then
			jump_time = jump_remaining;
		else
			jump_time = dt;
		end;
		
		air_time = air_time + dt;

		--[[
	local yVel = self:GetYVelocity();
	local angleInRadians = math.rad( self.angle );
	local invSinOfAngleInRadians = -math.sin( angleInRadians );
	local gSpeed = self.groundSpeed;
	if( gSpeed < 0 )then gSpeed = -gSpeed; end;
	local angleVel = ( gSpeed * invSinOfAngleInRadians );
	if( self.state == "on_ground" )then
		self:SetYVelocity( yVel + angleVel );
	end;
		--]]
		local yVel = self:GetYVelocity();
		--print(yVel);
		local jumpForce = self.jumpForce;
		self:SetYVelocity( yVel - (jumpForce * jump_time) );
		self.state = "jumping";	
	end;
	
	return( air_time );
--
end;
--]]
